from .plugin import MetaPlugin

from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg, NavigationToolbar2QT
from matplotlib import pyplot as plt

from .qt import *


class MatplotlibPlugin(MetaPlugin):

    def __init__(self, with_toolbar=False, with_zoom=False):

        self.plt = plt
        self.figure = self.plt.figure()

        layout = QVBoxLayout()
        self.widget = QWidget()
        self.widget.setLayout(layout)
        self.canvas = FigureCanvasQTAgg(self.figure)

        self.current_axis = None

        if with_toolbar:
            self.toolbar = NavigationToolbar2QT(self.canvas, self.canvas)
            layout.addWidget(self.toolbar)
            self.toolbar.pan()

        layout.addWidget(self.canvas)

        if with_zoom:
            def _zoom(e):
                factor = 2 if e.button == 'down' else 0.5
                xlim, ylim = self.current_axis.get_xlim(), self.current_axis.get_ylim()
                self.current_axis.set_xlim(
                    xlim[0] + (xlim[1] - xlim[0])*0.5 - (xlim[1] - xlim[0])*0.5*factor,
                    xlim[0] + (xlim[1] - xlim[0])*0.5 + (xlim[1] - xlim[0])*0.5*factor
                )
                self.current_axis.set_ylim(
                    ylim[0] + (ylim[1] - ylim[0])*0.5 - (ylim[1] - ylim[0])*0.5*factor,
                    ylim[0] + (ylim[1] - ylim[0])*0.5 + (ylim[1] - ylim[0])*0.5*factor
                )
                self.canvas.draw()
            self.canvas.mpl_connect('scroll_event', _zoom)

    def update(self):
        self.canvas.draw()

    def axis(self):
        self.current_axis = self.figure.add_subplot(111)
        return self.current_axis



from .plugin import MetaPlugin
from vispy import scene
from vispy.scene.visuals import Image


class VispyPlugin(MetaPlugin):
    def __init__(self):
        self.canvas = scene.SceneCanvas()
        self.view = self.canvas.central_widget.add_view()

        self.widget = self.canvas.native

    def add_pan_zoom_camera(self):
        self.view.camera = scene.PanZoomCamera(aspect=1)
        self.view.camera.flip = (0, 1, 0)

    def add_image(self, image_data):
        image = Image(data=image_data, parent=self.view.scene)
        if self.view.camera:
            self.view.camera.set_range()
        return image

    def before_update(self):
        pass

    def after_update(self):
        pass



from .qt import *


class Value(object):

    name = None
    label = None

    resulting_layout = None

    def _set_name_label(self, name, label):
        self.name = name
        if label is None:
            self.label = ' '.join(s.capitalize() for s in name.split('_'))
        else:
            self.label = label

    def generate_widget(self):
        raise RuntimeError('Virtual function call')

    def get_widget(self):
        if self.resulting_layout:
            return self.resulting_layout

        self.resulting_layout = self.generate_widget()

        return self.resulting_layout


class ListValue(Value):
    def __init__(self, value=None, choice=None, name="value", label=None):
        self._set_name_label(name, label)
        self.choice = list(choice)

        if value is None:
            value = choice[0]
        self.value = value

    def generate_widget(self):

        resulting_layout = QHBoxLayout()

        combobox = QComboBox()

        for item in self.choice:
            combobox.addItem(str(item))

        combobox.setCurrentIndex(self.choice.index(self.value))

        def index_changed():
            value_index = combobox.currentIndex()
            self.value = self.choice[value_index]

            if self.callback is not None:
                self.callback(self.name, self.value)

        combobox.currentIndexChanged.connect(index_changed)

        label = QLabel()
        label.setText(self.label)

        resulting_layout.addWidget(label)
        resulting_layout.addWidget(combobox)

        index_changed()

        return resulting_layout


class Action(Value):
    def __init__(self, name="value", label=None, callback=None):
        self._set_name_label(name, label)
        self.callback = callback
        self.value = False

    def generate_widget(self):

        resulting_layout = QHBoxLayout()

        pushbutton = QPushButton()

        pushbutton.setText(self.label)

        def pressed():
            if self.callback is not None:
                self.value = True
                self.callback(self.name, self.value)
                self.value = False

        pushbutton.pressed.connect(pressed)

        resulting_layout.addWidget(pushbutton)

        return resulting_layout


class Label(Value):
    def __init__(self, name="label", label=None, value=None):
        self._set_name_label(name, label)
        self.value = value

        self._label = label

    def update(self, value=None):
        if value is not None:
            self.value = value

        if self._label:
            if '%' in self.label:
                self._label.setText(self.label % (self.value,))
            else:
                self._label.setText(self.label)

    def generate_widget(self):
        resulting_layout = QHBoxLayout()
        self._label = QLabel()
        self.update()
        resulting_layout.addWidget(self._label)
        return resulting_layout


class NumericValue(Value):
    def __init__(self,
                 name="value", value=0.0, label=None, type_=float, minimum=0.0, maximum=1.0, callback=None, unit=""
                 ):
        self.value = value

        self._set_name_label(name, label)

        self.type_ = type_
        self.minimum = minimum
        self.maximum = maximum
        self.callback = callback

        self.unit = unit

        self.interval = None

        self.resulting_layout = None

    def format_label(self, value):
        return str(value)

    def generate_widget(self):

        resulting_layout = QHBoxLayout()

        slider = QSlider()

        textbox = QLineEdit()

        slider.setOrientation(Qt.Horizontal)

        rel_mi, rel_ma = 0, 1000
        abs_mi, abs_ma = self.minimum, self.maximum

        slider.setMinimum(rel_mi)
        slider.setMaximum(rel_ma)

        if self.interval:
            slider.setSingleStep(self.interval)

        slider.setMinimumHeight(10)

        rel_delta = rel_ma - rel_mi
        abs_delta = abs_ma - abs_mi

        def update(value):
            if value > abs_ma:
                value = abs_ma
            if value < abs_mi:
                value = abs_mi

            if self.interval:
                value = round((value - abs_mi)/self.interval) * self.interval + abs_mi

            rel_value = ((value-abs_mi) / abs_delta) * rel_delta + rel_mi

            slider.setValue(int(rel_value))

            value = self.type_(value)

            value_str = self.format_label(value)

            textbox.setText(value_str)

            self.value = value

            if self.callback is not None:
                self.callback(self.name, self.value)

        def slider_changed():
            value = slider.value()
            abs_value = ((value - rel_mi) / rel_delta) * abs_delta + abs_mi
            update(abs_value)

        def textbox_changed():
            value = float(textbox.text())
            update(value)

        slider.valueChanged.connect(slider_changed)
        textbox.textEdited.connect(textbox_changed)

        label = QLabel()
        label.setText(self.label)

        unit = QLabel()
        unit.setText(self.unit)

        resulting_layout.addWidget(label)
        resulting_layout.addWidget(slider)
        resulting_layout.addWidget(textbox)
        resulting_layout.addWidget(unit)

        update(self.value)

        return resulting_layout


class IntValue(NumericValue):
    def __init__(self, name="value", value=0.0, label=None, minimum=0.0, maximum=1.0, callback=None, unit="",
                 interval=1):
        super(IntValue, self).__init__(name, value, label, int, minimum, maximum, callback, unit)
        self.interval = interval


class FloatValue(NumericValue):
    def __init__(self, name="value", value=0.0, label=None, minimum=0.0, maximum=1.0, callback=None, digits=4, unit=""):
        super(FloatValue, self).__init__(name, value, label, float, minimum, maximum, callback, unit)
        self.digits = digits

    def format_label(self, value):
        return str(round(value, self.digits))


class Values(object):
    Value = Value
    IntValue = IntValue
    FloatValue = FloatValue
    ListValue = ListValue
    Label = Label
    Action = Action



import sys
from .qt import *


def get_argv_or_dialog():
    if len(sys.argv) > 1:
        return sys.argv[1:]
    else:
        filenames, _ = QFileDialog().getOpenFileNames()
        return filenames



import time

from yaval.plugin import MetaPlugin
from yaval.values import Values

from .qt import *


class Visualizer(QDialog):

    title = "Visualizer"

    result_table = False
    print_update_time = False

    Values = Values

    def __init__(self, parent=None):

        super(Visualizer, self).__init__(parent)
        self.plugin = MetaPlugin()
        self.setWindowTitle(self.title)

        if False:
            menu_bar = QMenuBar()

            menu_file = menu_bar.addMenu("File")

            open_action = menu_file.addAction("Open")

            def _dummy():
                print("CLICKED")
            open_action.triggered.connect(_dummy)
            print(menu_file.addAction("Exit"))

        outer_layout = QVBoxLayout()

        layout = QHBoxLayout()
        value_container = QVBoxLayout()
        splitter = QSplitter()

        layout.addWidget(splitter)

        self.values = []
        self.value_input_target = value_container

        self.setLayout(outer_layout)

        # outer_layout.addWidget(menu_bar)
        outer_layout.addLayout(layout)

        # layout.addWidget(canvas.native)
        # layout.addLayout(value_container)

        receptor_widget = QWidget()
        receptor = QVBoxLayout()
        receptor_widget.setLayout(receptor)

        self.receptor = receptor

        # receptor.addWidget()

        # receptor.addWidget(canvas.native)

        splitter.addWidget(receptor_widget)

        empty = QWidget()
        empty.setLayout(value_container)

        splitter.addWidget(empty)

        model = ListOfDictsModel()

        model.update_data([
            #    {"a": 2, "b": 3, "c": 4},
            #    {"a": 12, "b": 23, "c": 54},
            #    {"a": 22, "b": 33, "c": 764},
            #    {"a": 32, "b": 43, "c": 74}
        ])

        # QApplication.clipboard().setText(model.get_clipboard_str())

        self.output_model = model

        tv = QTableView()
        tv.setModel(model)
        tv.setEditTriggers(QAbstractItemView.NoEditTriggers)

        result_layout = QVBoxLayout()

        result_layout.addWidget(tv)

        copy_button = QPushButton()
        copy_button.setText("Copy to Clipboard")

        def _copy():
            QApplication.clipboard().setText(model.get_clipboard_str())

        copy_button.clicked.connect(_copy)

        result_layout.addWidget(copy_button)

        result_layout_widget = QWidget()
        result_layout_widget.setLayout(result_layout)

        if self.result_table:
            splitter.addWidget(result_layout_widget)

        self.visualization_callback = self.visualization()

        self.plugin.before_update()
        self.visualization_callback(self.get_values())
        self.plugin.after_update()

    def register_plugin(self, plugin):
        self.receptor.addWidget(plugin.get_widget())
        self.plugin = plugin
        return plugin

    def visualization_callback(self, values):
        pass

    # noinspection PyUnusedLocal
    def value_change_callback(self, what, to):
        before = time.time()
        values = self.get_values()
        self.plugin.before_update()
        self.visualization_callback(values)
        self.plugin.after_update()
        elapsed = time.time() - before
        if self.print_update_time:
            print("whole update took %.4fs" % (elapsed,))

    def add_values(self, *values):
        for value in values:
            if isinstance(value, Values.Value):
                self.add_value(value)
            else:
                for inner_value in value:
                    self.add_value(inner_value)

    def add_value(self, value):
        if value not in self.values:
            value.callback = self.value_change_callback
            self.values.append(value)
            self.value_input_target.addLayout(value.get_widget())

    def get_values(self):

        class AccessByMemberDict(dict):
            def __getattr__(self, item):
                if item in self:
                    return self[item]

        return AccessByMemberDict({value.name: value.value for value in self.values})

    def visualization(self):
        def _vis(_):
            pass
        return _vis

    @classmethod
    def run(cls):
        import vispy
        app = vispy.app.use_app('PySide2')
        _ = app.native
        c_ = cls()
        c_.show()
        app.run()


class ListOfDictsModel(QStandardItemModel):
    _data = None
    _header = None

    def __init__(self):
        super(ListOfDictsModel, self).__init__()

    def update_data(self, data, header=None):

        self._data = data

        if header is None:
            try:
                self._header = list(sorted(data[0].keys()))
            except IndexError:
                self._header = []
        else:
            self._header = header

        self.setColumnCount(len(self._header))
        self.setRowCount(len(self._data))

        for n, header_item in enumerate(self._header):
            self.setHorizontalHeaderItem(n, QStandardItem(str(header_item)))

        for row, row_items in enumerate(self._data):
            for n, header_item in enumerate(self._header):
                self.setItem(row, n, QStandardItem(str(row_items[header_item])))

    def get_clipboard_str(self):
        result = ""

        result += "\t".join(str(h) for h in self._header) + "\n"

        for row, row_items in enumerate(self._data):
            result += "\t".join(str(row_items[h]) for h in self._header) + "\n"

        return result



# -*- coding: utf-8 -*-
"""
documentation
"""

from setuptools import setup, find_packages


setup(
    name='yaval',
    version='0.0.1-1',
    description='yet another visualization abstraction layer',
    long_description='',
    author='Christian C. Sachs',
    author_email='sachs.christian@gmail.com',
    url='',
    packages=find_packages(),
    requires=['numpy', 'matplotlib', 'vispy', 'pyside2'],
    license='BSD',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Operating System :: POSIX',
        'Operating System :: POSIX :: Linux',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Image Recognition',
    ]
)



from PySide2.QtWidgets import (
    QApplication,
    QDialog, QMenuBar, QVBoxLayout, QHBoxLayout, QSplitter,
    QWidget,
    QPushButton,
    QTableView,
    QAbstractItemView,
    QSlider,
    QLineEdit,
    QLabel,
    QComboBox,
    QFileDialog,  # keep it in
)

from PySide2.QtGui import (
    QStandardItemModel,
    QStandardItem,
)

from PySide2.QtCore import Qt



# by christian c. sachs

# correctly prime everything for PySide

import matplotlib
import vispy.app

matplotlib.use('Qt5Agg')
matplotlib.rcParams['backend.qt5'] = 'PySide2'

_ = vispy.app.use_app(backend_name='PySide2')

# noinspection PyPep8
from .plugin_matplotlib import MatplotlibPlugin
# noinspection PyPep8
from .plugin_vispy import VispyPlugin
# noinspection PyPep8
from .values import Values
# noinspection PyPep8
from .visualizer import Visualizer
